/*

kHTTPd -- the next generation

Send actual file-data to the connections

*/
/****************************************************************
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************/

/*

Purpose:

DataSending does the actual sending of file-data to the socket.

Note: Since asynchronous reads do not -yet- exists, this might block!

Return value:
	The number of requests that changed status
*/

#include <linux/kernel.h>
#include <linux/skbuff.h>

#include <net/tcp.h>

#include <asm/uaccess.h>
#include <linux/smp_lock.h>

#include "structure.h"
#include "prototypes.h"
#include "config.h"
static	char	*Block[KHTTPD_NUMCPU];

int DataSending(int CPU)
{
	struct http_request *CurrentRequest,**Prev;
	int count = 0;
	
	
	Prev = &(DataSendingQueue[CPU]);
	CurrentRequest = DataSendingQueue[CPU];
	while (CurrentRequest!=NULL)
	{
#ifndef SENDFILE	
		int ReadSize;
		int retval;
		mm_segment_t oldfs;


		
		ReadSize = min(4096,CurrentRequest->FileLength - CurrentRequest->BytesSent);
		
		if (ReadSize>0)
		{
			CurrentRequest->filp->f_pos = CurrentRequest->BytesSent;
		
			lock_kernel();
			oldfs = get_fs(); set_fs(KERNEL_DS);
			retval = CurrentRequest->filp->f_op->read(CurrentRequest->filp, Block[CPU], ReadSize, &CurrentRequest->filp->f_pos);
			set_fs(oldfs);
			unlock_kernel();
		
			if (retval>0)
			{
				retval = SendBuffer_async(CurrentRequest->sock,Block[CPU],retval);
				if (retval>0)
				{
					CurrentRequest->BytesSent += retval;
					count++;
				
				}
			}
		
		}
#endif				
		if ((CurrentRequest->BytesSent>=CurrentRequest->FileLength)||
		   (CurrentRequest->sock->sk->state !=TCP_ESTABLISHED))
		{
			struct http_request *Next;
			Next = CurrentRequest->Next;
			
			
			if (CurrentRequest->sock->sk->state ==TCP_ESTABLISHED)
			{
				lock_sock(CurrentRequest->sock->sk);
				CurrentRequest->sock->sk->nonagle = 0;
				CurrentRequest->sock->sk->linger = 1;
				tcp_push_pending_frames(CurrentRequest->sock->sk,&(CurrentRequest->sock->sk->tp_pinfo.af_tcp));
				release_sock(CurrentRequest->sock->sk);
			}	


			(*Prev) = CurrentRequest->Next;
			
			CurrentRequest->Next = LoggingQueue[CPU];
			LoggingQueue[CPU] = CurrentRequest;	
				
			CurrentRequest = Next;
			continue;
		
		}
		

		Prev = &(CurrentRequest->Next);	
		CurrentRequest = CurrentRequest->Next;
	}
	
	return count;
}

int InitDataSending()
{
	int I,I2;
	
	I=0;
	while (I<KHTTPD_NUMCPU)
	{
		Block[I] = (char*)get_free_page(GFP_KERNEL);
		if (Block[I] == NULL) 
		{
			I2=0;
			while (I2<KHTTPD_NUMCPU-1)
			{
				free_page((unsigned long)Block[I2++]);
			}
			return -1;
		};
		I++;
	}
	return 0;		
}

void StopDataSending(int CPU)
{
	struct http_request *CurrentRequest;
	CurrentRequest = DataSendingQueue[CPU];

	while (CurrentRequest!=NULL)
	{	
		CleanUpRequest(CurrentRequest);
		CurrentRequest=CurrentRequest->Next;		
	}

	free_page( (unsigned long)Block[CPU]);
}