/*

kHTTPd -- the next generation

Main program


kHTTPd TNG consists of 1 thread, this main-thread handles ALL connections
simultanious. It does this by keeping queues with the requests in different
stages.

The stages are

<not accepted> 		-	TCP/IP connection is not accepted yet
WaitForHeaders		-	Connection is accepted, waiting for headers
DataSending		-	Headers decoded, sending file-data
Userspace		-	Requires userspace daemon 
Logging			-	The request is finished, cleanup and logging

A typical flow for a request would be:

<not accepted>
WaitForHeaders
DataSending
Logging

or

<not accepted>
WaitForHeaders
Userspace
Logging



*/
/****************************************************************
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/wait.h>


#include "structure.h"
#include "prototypes.h"
#include "config.h"

struct http_request *WaitForHeaderQueue[KHTTPD_NUMCPU];
struct http_request *DataSendingQueue[KHTTPD_NUMCPU];
struct http_request *LoggingQueue[KHTTPD_NUMCPU];
struct http_request *UserspaceQueue[KHTTPD_NUMCPU];

char CurrentTime[64];

atomic_t	ConnectCount;

int ConnectionsPending(int CPU);

struct task_struct *tsk[KHTTPD_NUMCPU];
struct wait_queue main_wait[KHTTPD_NUMCPU];

static struct wait_queue  *DummyWQ[KHTTPD_NUMCPU];

int MainDaemon(void *cpu_pointer)
{
	int CPU;
	sigset_t tmpsig;
	
	
	MOD_INC_USE_COUNT;

	CPU=0;
	if (cpu_pointer!=NULL)
	CPU=(int)*(int*)cpu_pointer;

	exit_mm(current);
	sprintf(current->comm,"khttpd - %i",CPU);
	

	/* Block all signals except SIGKILL, SIGSTOP and SIGHUP */
	spin_lock_irq(&current->sigmask_lock);
	tmpsig = current->blocked;
	siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGSTOP)| sigmask(SIGHUP));
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
	
	
	tsk[CPU] = current;
	
	main_wait[CPU].task = tsk[CPU];
	main_wait[CPU].next = NULL;
	
	add_wait_queue(MainSocket->sk->sleep,&(main_wait[CPU]));
	
	
	while (sysctl_khttpd_stop==0)
	{
		int changes = 0;
				

		
		changes +=AcceptConnections(CPU,MainSocket,O_NONBLOCK);
		changes +=WaitForHeaders(CPU);
		changes +=DataSending(CPU);
		changes +=Userspace(CPU);
		changes +=Logging(CPU);
		
		current->state = TASK_INTERRUPTIBLE;	
		if (changes==0) 
		{
			interruptible_sleep_on_timeout(&(DummyWQ[CPU]),10);	
			if (CPU==0) 
				UpdateCurrentDate();
		}
			
		if (signal_pending(current))
			break;		  
	
	}
	
	StopListening();
	
	StopWaitingForHeaders(CPU);
	StopDataSending(CPU);
	StopUserspace(CPU);
	StopLogging(CPU);
	MOD_DEC_USE_COUNT;
	printk("Ended kHTTPd\n");
	return 0;
}

int CountBuf[KHTTPD_NUMCPU];

int init_module(void)
{
	int I;
	
	I=0;
	while (I<KHTTPD_NUMCPU)
	{
		CountBuf[I]=I++;
	}
	memset(WaitForHeaderQueue, 0, sizeof(WaitForHeaderQueue));
	memset(DataSendingQueue	 , 0, sizeof(DataSendingQueue));
	memset(LoggingQueue	 , 0, sizeof(LoggingQueue));
	memset(UserspaceQueue    , 0, sizeof(UserspaceQueue));
	memset(DummyWQ		 , 0, sizeof(DummyWQ));
	
	atomic_set(&ConnectCount,0);
	
	if (InitDataSending()!=0)
	   return -1;
	if (InitWaitHeaders()!=0)
	   return -1;
	   
	if (!StartListening(KHTTPD_PORT))
		return -1;
		
 	AddMimeType(".htm","text/html");
 	AddMimeType("html","text/html");
 	AddMimeType(".gif","image/gif");
 	AddMimeType(".jpg","image/jpeg");
 	AddMimeType(".png","image/png");
 	AddMimeType(".zip","application/zip");
 	AddMimeType("r.gz","application/x-gtar");
 	AddMimeType(".tgz","application/x-gtar");
	AddMimeType(".deb","application/x-debian");
	AddMimeType("lass","application/x-java");
	AddMimeType(".txt","text/plain");
	
	AddDynamicString("..");
	AddDynamicString("cgi-bin");
	
	StartSysctl();
		
	
	I=0;
	while (I<KHTTPD_NUMCPU)
		kernel_thread(MainDaemon,&(CountBuf[I++]),0);
#ifdef SENDFILE	
	kernel_thread(Sendfile_Daemon,NULL,0);
#endif
	
	return 0;
}

void cleanup_module(void)
{
	EndSysctl();
}

int ConnectionsPending(int CPU)
{
	if (WaitForHeaderQueue[CPU]!=NULL) return O_NONBLOCK;
	if (DataSendingQueue[CPU]!=NULL) return O_NONBLOCK;
	if (LoggingQueue[CPU]!=NULL) return O_NONBLOCK;
	if (UserspaceQueue[CPU]!=NULL) return O_NONBLOCK;
  return O_NONBLOCK;
}

void PrintStatus(int CPU)
{
	printk("Status of khttpd\n");
	printk("------------------\n");
	if (WaitForHeaderQueue[CPU]!=NULL) printk("WaitForHeaderQueue \n");
	if (DataSendingQueue[CPU]!=NULL) printk("DataSendingQueue \n");
	if (LoggingQueue[CPU]!=NULL) printk("LoggingQueue \n");
	if (UserspaceQueue[CPU]!=NULL) printk("UserspaceQueue \n");
	printk("------------------\n");

}