/*

kHTTPd -- the next generation

General functions

*/
/****************************************************************
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************/

#include <linux/kernel.h>

#include <linux/ctype.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/net.h>
#include <linux/sched.h>
#include <linux/skbuff.h>
#include <linux/unistd.h>
#include <linux/file.h>
#include <linux/smp_lock.h>

#include <net/ip.h>
#include <net/sock.h>

#include <asm/atomic.h>
#include <asm/semaphore.h>
#include <asm/processor.h>
#include <asm/uaccess.h>

#include "structure.h"
#include "prototypes.h"

void CleanUpRequest(struct http_request *Req)
{
	
	if (Req->sock!=NULL)
	{
		remove_wait_queue(Req->sock->sk->sleep,&(Req->sleep));
	    	sock_release(Req->sock);
	}
	
	if (Req->filp!=NULL)
	{
		lock_kernel();
	    	fput(Req->filp);
		unlock_kernel();
	}
	
	kfree(Req);
	
	atomic_dec(&ConnectCount);
}

int SendBuffer(struct socket *sock, const char *Buffer, size_t Length)
{
	struct msghdr	msg;
	mm_segment_t	oldfs;
	struct iovec	iov;
	int 		len;
	
	msg.msg_name     = 0;
	msg.msg_namelen  = 0;
	msg.msg_iov	 = &iov;
	msg.msg_iovlen   = 1;
	msg.msg_control  = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags    = 0;    
	iov.iov_base = (char*) Buffer;
	iov.iov_len  = (size_t)Length;
	
	
	len = 0;

	oldfs = get_fs(); set_fs(KERNEL_DS);
	len = sock_sendmsg(sock,&msg,(size_t)(Length-len));
	set_fs(oldfs);


	return len;	
}

int SendBuffer_async(struct socket *sock, const char *Buffer, size_t Length)
{
	struct msghdr	msg;
	mm_segment_t	oldfs;
	struct iovec	iov;
	int 		len;
	
	msg.msg_name     = 0;
	msg.msg_namelen  = 0;
	msg.msg_iov	 = &iov;
	msg.msg_iovlen   = 1;
	msg.msg_control  = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags    = MSG_DONTWAIT;    
	iov.iov_base = (char*) Buffer;
	iov.iov_len  = Length;
	
	
	oldfs = get_fs(); set_fs(KERNEL_DS);
	len = sock_sendmsg(sock,&msg,(size_t)(Length));
	set_fs(oldfs);
	
	return len;	
}


void ReadRest(struct socket *sock)
{
	struct msghdr		msg;
	struct iovec		iov;
	static char		Buffer[1024];
	int			len;

	mm_segment_t		oldfs;
	
	
	len = 1;
	
	if (sock->sk->state!=TCP_ESTABLISHED)
		return;
		
	while (len>0)
	{
	/* First, read the data */
	
	msg.msg_name     = 0;
	msg.msg_namelen  = 0;
	msg.msg_iov	 = &iov;
	msg.msg_iovlen   = 1;
	msg.msg_control  = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags    = MSG_DONTWAIT;
	
	iov.iov_base = &Buffer[0];
	iov.iov_len  = (size_t)1024;
	
	len = 0;

	oldfs = get_fs(); set_fs(KERNEL_DS);
	len = sock_recvmsg(sock,&msg,4096,MSG_DONTWAIT);
	set_fs(oldfs);
	}
}


char NoPerm[] = "403 Permission denied\r\nServer: kHTTPd 0.1.4\r\n\r\n Permission denied";
char NotModified[] = "304 Not Modified\r\nServer: kHTTPd 0.1.4\r\n\r\n";


void Send403(struct socket *sock)
{
	SendBuffer(sock,NoPerm,strlen(NoPerm));
}

void Send304(struct socket *sock)
{
	SendBuffer(sock,NotModified,strlen(NoPerm));
}

