#ifndef _INCLUDE_GUARD_PROTOTYPES_H
#define _INCLUDE_GUARD_PROTOTYPES_H


#include <linux/kernel.h>
#include <linux/net.h>
#include <linux/time.h>
#include <linux/wait.h>
#include <net/sock.h>
#include <asm/uaccess.h>

#include "structure.h"
#include "config.h"

/* sockets.c */
int  StartListening(int Port);
void StopListening(void);

extern struct socket *MainSocket;


/* sysctl.c */
void StartSysctl(void);
void EndSysctl(void);

extern int sysctl_khttpd_stop;


/* main.c */
extern struct http_request *WaitForHeaderQueue[KHTTPD_NUMCPU];
extern struct http_request *DataSendingQueue[KHTTPD_NUMCPU];
extern struct http_request *LoggingQueue[KHTTPD_NUMCPU];
extern struct http_request *UserspaceQueue[KHTTPD_NUMCPU];
extern char CurrentTime[];
extern atomic_t ConnectCount;
extern struct wait_queue main_wait[KHTTPD_NUMCPU];
void PrintStatus(int CPU);

/* misc.c */

void CleanUpRequest(struct http_request *Req);
int SendBuffer(struct socket *sock, const char *Buffer, size_t Length);
int SendBuffer_async(struct socket *sock, const char *Buffer, size_t Length);
void ReadRest(struct socket *sock);
void Send403(struct socket *sock);
void Send304(struct socket *sock);

/* accept.c */

int AcceptConnections(int CPU,struct socket *Socket,int Block);

/* waitheaders.c */

int WaitForHeaders(int CPU);
void StopWaitingForHeaders(int CPU);
int DecodeHeader(int CPU,struct http_request *Request);
int InitWaitHeaders(void);



/* datasending.c */

int DataSending(int CPU);
void StopDataSending(int CPU);
int InitDataSending(void);


/* userspace.c */

int Userspace(int CPU);
void StopUserspace(int CPU);
int AddSocketToAcceptQue(struct socket *sock,int Port);


/* rfc_time.c */

void time_Unix2RFC(time_t Zulu,char *Buffer);
void UpdateCurrentDate(void);
time_t mimeTime_to_UnixTime(char *Q);
extern int CurrentTime_i;

/* rfc.c */

void ParseHeader(char *Buffer,int length, struct http_request *Head);
void HTTPHeader(char *Filename,int FileSize,time_t MTime, char *Buf);
void AddMimeType(char *Ident,char *Type);
void SendHTTPHeader(struct http_request *Request);



/* security.c */
struct file *OpenFileForSecurity(char *Filename);
void AddDynamicString(char *String);
void GetSecureString(char *String);

/* logging.c */
int Logging(int CPU);
void StopLogging(int CPU);

/* sendfile.c */
int Sendfile_Daemon(void* unused);
int AddSendFile(struct async_sendfile *Data, struct wait_queue *WQ);


#endif