/*

kHTTPd -- the next generation

RFC related functions (headers and stuff)

*/

/****************************************************************
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************/


#include <linux/kernel.h>

#include <linux/ctype.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/net.h>
#include <linux/sched.h>
#include <linux/skbuff.h>
#include <linux/unistd.h>
#include <linux/file.h>

#include <net/ip.h>
#include <net/sock.h>

#include <asm/atomic.h>
#include <asm/semaphore.h>
#include <asm/processor.h>
#include <asm/uaccess.h>
#include <asm/msr.h>


#include "prototypes.h"
#include "structure.h"
#include "config.h"

atomic_t	MimeCount;

struct MimeType
{
	__u32 	identifier;
	char	type[32-4-sizeof(int)];  
	int	len;
};

static struct MimeType	MimeTypes[KHTTPD_NUMMIMETYPES];


void AddMimeType(char *Ident,char *Type)
{	
	__u32	*I;
	
	if (strlen(Ident)!=4) 
   	{	
   		printk("httpd: Only 4-byte mime-identifiers are accepted\n");
   		return;
   	}

	if (strlen(Type)>(32-4-sizeof(int))) 
   	{	
   		printk("httpd: Mime-string too long.\n");
   		return;
   	}
   	
   	I=(__u32*)Ident;
   	
   	/* FIXME: Need to lock-down all access to the mime-structure here */
   	
   	MimeTypes[atomic_read(&MimeCount)].identifier=*I;
   	strcpy(MimeTypes[atomic_read(&MimeCount)].type,Type);
   	MimeTypes[atomic_read(&MimeCount)].len = strlen(Type);
   	
   	atomic_inc(&MimeCount);
}


__inline char *ResolveMimeType(char *File)
/*

	The returned string is for READ ONLY!

*/
{	
	__u32	*I;
	int pos,lc,filelen;
	
	filelen = strlen(File);
	
	if (filelen<4) 
   	{	
   		return NULL;
   	}

	pos = filelen-4;
   	I=(__u32*)(File+pos);
   	
   	lc=0;
   	
   	while (lc<atomic_read(&MimeCount))
 	{
   		if (MimeTypes[lc].identifier == *I)
   	  		return MimeTypes[lc].type;
   	  	lc++;
   	} 	
   	return MimeTypes[0].type;
}

__inline char *ResolveMimeType2(char *File,int *Len)
/*

	The returned string is for READ ONLY!

*/
{	
	__u32	*I;
	int pos,lc,filelen;
	
	*Len = 0;
	
	filelen = strlen(File);
	
	if (filelen<4) 
   	{	
   		return NULL;
   	}

	pos = filelen-4;
   	I=(__u32*)(File+pos);
   	
   	lc=0;
   	
   	while (lc<atomic_read(&MimeCount))
 	{
   		if (MimeTypes[lc].identifier == *I)
   		{
   			*Len = MimeTypes[lc].len;
   	  		return MimeTypes[lc].type;
   	  	}
   	  	lc++;
   	} 	
   	
	*Len = MimeTypes[0].len;  
   	return MimeTypes[0].type;
}

void HTTPHeader(char *Filename,int FileSize,time_t MTime, char *BUF)
{
	char TimeBuf[64];


#ifndef BENCHMARK     
		time_Unix2RFC(MTime,TimeBuf);
	sprintf(BUF,
	  "HTTP/1.0 200 OK\r\nServer: kHTTPd/0.1.4\r\nDate: %s\r\nContent-type: %s\r\nLast-modified: %s\r\nContent-length: %i\r\n\r\n",
	  CurrentTime,ResolveMimeType(Filename),TimeBuf,FileSize);
#else	
	sprintf(BUF,
	  "HTTP/1.0 200 OK\r\nContent-type: text/html\r\nContent-length: %i\r\n\r\n",
	  FileSize);
#endif
}

char HeaderPart1[] = "HTTP/1.0 200 OK\r\nServer: kHTTPd/0.1.4\r\nDate: ";
char HeaderPart1b[] ="HTTP/1.0 200 OK";
char HeaderPart3[] = "\r\nContent-type: ";
char HeaderPart5[] = "\r\nLast-modified: ";
char HeaderPart7[] = "\r\nContent-length: ";
char HeaderPart9[] = "\r\n\r\n";

#ifdef BENCHMARK
void SendHTTPHeader(struct http_request *Request)
{
	struct msghdr	msg;
	mm_segment_t	oldfs;
	struct iovec	iov[9];
	int 		len,len2;
	
	

		
	msg.msg_name     = 0;
	msg.msg_namelen  = 0;
	msg.msg_iov	 = &iov[0];
	msg.msg_iovlen   = 6;
	msg.msg_control  = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags    = 0;  /* Synchronous for now */
	
	iov[0].iov_base = HeaderPart1b;
	iov[0].iov_len  = 15;
	iov[1].iov_base = HeaderPart3;
	iov[1].iov_len  = 16;
	iov[2].iov_base = ResolveMimeType2(Request->FileName,&(iov[2].iov_len));
	
	iov[3].iov_base = HeaderPart7;
	iov[3].iov_len  = 18;
	
	
	sprintf(Request->LengthS,"%i",Request->FileLength);
	iov[4].iov_base = Request->LengthS;
	iov[4].iov_len  = strlen(Request->LengthS);
	iov[5].iov_base = HeaderPart9;
	iov[5].iov_len  = 4;
	
	len2=15+16+18+iov[2].iov_len+iov[4].iov_len+4;
	
	
	len = 0;
	

	oldfs = get_fs(); set_fs(KERNEL_DS);
	len = sock_sendmsg(Request->sock,&msg,len2);
	set_fs(oldfs);


	return;	
}
#else
void SendHTTPHeader(struct http_request *Request)
{
	struct msghdr	msg;
	mm_segment_t	oldfs;
	struct iovec	iov[9];
	int 		len,len2;
	
		
	msg.msg_name     = 0;
	msg.msg_namelen  = 0;
	msg.msg_iov	 = &iov[0];
	msg.msg_iovlen   = 9;
	msg.msg_control  = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags    = 0;  /* Synchronous for now */
	
	iov[0].iov_base = HeaderPart1;
	iov[0].iov_len  = 45;
	iov[1].iov_base = CurrentTime;
	iov[1].iov_len  = 29;
	iov[2].iov_base = HeaderPart3;
	iov[2].iov_len  = 16;
	iov[3].iov_base = ResolveMimeType2(Request->FileName,&(iov[3].iov_len));
	
	iov[4].iov_base = HeaderPart5;
	iov[4].iov_len  = 17;
	time_Unix2RFC(min(Request->Time,CurrentTime_i),Request->TimeS);
	  /* The min() is required by rfc1945, section 10.10 */
	
	iov[5].iov_base = Request->TimeS;
	iov[5].iov_len  = 29;
	iov[6].iov_base = HeaderPart7;
	iov[6].iov_len  = 18;
	
	
	sprintf(Request->LengthS,"%i",Request->FileLength);
	iov[7].iov_base = Request->LengthS;
	iov[7].iov_len  = strlen(Request->LengthS);
	iov[8].iov_base = HeaderPart9;
	iov[8].iov_len  = 4;
	
	len2=45+2*29+16+17+18+iov[7].iov_len+4+iov[3].iov_len;
	
	len = 0;
	
	oldfs = get_fs(); set_fs(KERNEL_DS);
	len = sock_sendmsg(Request->sock,&msg,len2);
	set_fs(oldfs);


	return;	
}
#endif


void ParseHeader(char *Buffer,int length, struct http_request *Head)
{
	char *Endval,*EOL,*tmp;
	
	Endval = Buffer + length;
	
/*	memset(Head->FileName,0,sizeof(Head->FileName));
	memset(Head->IMS,0,sizeof(Head->IMS));
	memset(Head->Agent,0,sizeof(Head->Agent));
	memset(Head->Host,0,sizeof(Head->Host));
*/	
	
	while (Buffer<Endval)
	{
		if (isspace(Buffer[0]))
		{
			Buffer++;
			continue;
		};
			
		
		EOL=strchr(Buffer,'\n');
		
		if (EOL==NULL) EOL=Endval;
		
		if (EOL-Buffer<4) 
		{
			Buffer++;
			continue;
		};
		
		if (strncmp("GET ",Buffer,4)==0)
		{
			int PrefixLen;
			Buffer+=4;
			
			tmp=strchr(Buffer,' ');
			if (tmp==0) tmp=EOL-1;
			
			if (tmp>Endval) continue;
			
			strcpy(Head->FileName,KHTTPD_PREFIX);
			PrefixLen = strlen(KHTTPD_PREFIX);
			Head->FileNameLength = min(255,tmp-Buffer+PrefixLen);		
			
			strncat(Head->FileName,Buffer,min(255-PrefixLen,tmp-Buffer));
					
			Buffer=EOL+1;	
#ifdef BENCHMARK
			break;
#endif						
			continue;
		}
#ifndef BENCHMARK		
		if (strncmp("If-Modified-Since: ",Buffer,19)==0)
		{
			Buffer+=19;
			
			strncpy(Head->IMS,Buffer,min(128,EOL-Buffer-1));
					
			Buffer=EOL+1;	
			continue;
		}

		if (strncmp("User-Agent: ",Buffer,12)==0)
		{
			Buffer+=12;
			
			strncpy(Head->Agent,Buffer,min(128,EOL-Buffer-1));
					
			Buffer=EOL+1;	
			continue;
		}
		

		if (strncmp("Host: ",Buffer,6)==0)
		{
			Buffer+=6;
			
			strncpy(Head->Host,Buffer,min(128,EOL-Buffer-1));
					
			Buffer=EOL+1;	
			continue;
		}
#endif		
		Buffer = EOL+1;  /* Skip line */
	}

}