#ifdef SENDFILE
/*

kHTTPd -- the next generation

Asynchronous sendfile daemon

*/
/****************************************************************
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************/

/*

*/

#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <net/tcp.h>
#include <asm/uaccess.h>
#include <linux/malloc.h>
#include <linux/shm.h>
#include <linux/mman.h>
#include <linux/locks.h>
#include <linux/pagemap.h>
#include <linux/swap.h>
#include <linux/smp_lock.h>
#include <linux/blkdev.h>
#include <linux/file.h>
#include <linux/swapctl.h>
#include <linux/slab.h>
#include <linux/mm.h>

#include <asm/pgtable.h>

#include "structure.h"
#include "prototypes.h"


static struct async_sendfile *SendfileQueue=NULL; 
static struct semaphore async_sendfile_mutex = MUTEX;
static struct wait_queue  *DummyWQ=NULL;

static struct  task_struct *SendfileD;

static int TryFile(struct async_sendfile *File);


int Sendfile_Daemon(void* unused)
{
	struct async_sendfile *Current,**Prev,*Temp;
	int changes = 0;
	
	exit_mm(current);
	sprintf(current->comm,"kasyncd");
	
	SendfileD = current;
	

	
	while (sysctl_khttpd_stop == 0)
	{
		down(&async_sendfile_mutex);
		
		Current = SendfileQueue;
		Prev = &SendfileQueue;
		
		changes = 0;
		while (Current!=NULL)
		{
/*			printk("Pos = %i,   Length = %i \n",*(Current->Pos),*(Current->Length));*/

			if (*(Current->Pos) >= *(Current->Length)||(Current->sock->sk->state!=TCP_ESTABLISHED))
			{ /* Request is done, dequeue it */
				*Prev = Current->Next;
				Temp = Current->Next;
				if (Current->page_cache)
				 	page_cache_free(Current->page_cache);
				 	
				if (!Current->clientsleep)
				  wake_up_process(Current->clientsleep->task);
				kfree(Current);
				Current = Temp;
				continue;
			}
			changes += TryFile(Current);
			
			Prev = &(Current->Next);
			Current = Current->Next;
		}
		up(&async_sendfile_mutex);
		
		current->state = TASK_INTERRUPTIBLE;	
		if (changes==0) 
			interruptible_sleep_on_timeout(&DummyWQ,1);	

	}
	
	
	printk("Ended asynchronous Sendfile daemon \n");
	return 0;
}

int AddSendFile(struct async_sendfile *Data, struct wait_queue *WQ)
{
	down(&async_sendfile_mutex);
	
	Data->Next = SendfileQueue;
	SendfileQueue = Data;
	
	Data->inode = Data->filp->f_dentry->d_inode;
	
	Data->sleep.task = SendfileD;
	Data->clientsleep = WQ;
	
	up(&async_sendfile_mutex);
	
	wake_up_process(SendfileD);	
	
	return 0; /* Should return "ticket number" later */
}

static inline int get_max_readahead(struct inode * inode)
{
	if (!inode->i_dev || !max_readahead[MAJOR(inode->i_dev)])
		return MAX_READAHEAD;
	return max_readahead[MAJOR(inode->i_dev)][MINOR(inode->i_dev)];
}

static inline void add_to_page_cache(struct page * page,
	struct inode * inode, unsigned long offset,
	struct page **hash)
{
	atomic_inc(&page->count);
	page->flags = (page->flags & ~((1 << PG_uptodate) | (1 << PG_error))) | (1 << PG_referenced);
	page->offset = offset;
	add_page_to_inode_queue(inode, page);
	__add_page_to_hash_queue(page, hash);
}


static int TryFile(struct async_sendfile *File)
{
	size_t	pos;
	int	error,offset,size,size2;
	int	WorkDone = 0;


/*	printk("TryFile \n");*/
	pos = *(File->Pos);
	
	/* Step 1: If we haven't found the page, try to find it now */
/*	printk("TryFile: Step 1 \n");*/
	
/*	if (!File->page)	*/
	{
		File->hash = page_hash(File->inode,pos & PAGE_CACHE_MASK);
		File->page = __find_page(File->inode,pos & PAGE_CACHE_MASK, *(File->hash));
	}
	
	if (!File->page)  /* We can't find the page, so we have to allocate it */
	{
		/* Step 2: Allocate a spare page */
/*		printk("TryFile: Step 2 \n");*/
		if (!File->page_cache)
		{
			File->page_cache = page_cache_alloc();
			/* This may have slept. We should start all over; we have
			   made some progress so we return positive */
			return 1;
		}
		
		/* Step 3: Add the page to the page-cache */
/*		printk("TryFile: Step 3 \n");*/
		
		File->page = page_cache_entry(File->page_cache);
		File->page_cache = 0;
		add_to_page_cache(File->page,File->inode,pos & PAGE_CACHE_MASK,File->hash);
	
		/* Step 4: Issue the read-request on the page */
/*		printk("TryFile: Step 4 \n");*/
		
		File->filp->f_ramax = get_max_readahead(File->inode);		   
		error = File->inode->i_op->readpage(File->filp,File->page);
		if (error) goto HaveError;
		WorkDone++;
		}
	/* Step 5: Wait for the page to become available and valid */
/*	printk("TryFile: Step 5 \n");*/
	
	if (PageLocked(File->page))
	{
		if (!File->page->wait)
		  add_wait_queue(&(File->page->wait),&(File->sleep));
		return WorkDone;   
	}
	
	if (!PageUptodate(File->page)) /* Woops.... re-read it */
	{
/*		printk("TryFile: Step 5 - reread \n");*/
		error = File->inode->i_op->readpage(File->filp,File->page);
		if (error) goto HaveError;
		return WorkDone;	
	}
	
	/* Step 6: Send the data */
/*	printk("TryFile: Step 6 \n");*/
	
	offset = pos & ~PAGE_CACHE_MASK;
	size   = PAGE_CACHE_SIZE - offset;
	size = min(size, File->inode->i_size - pos);
	
	size2 = SendBuffer_async(File->sock,(const char*) (page_address(File->page)+offset),size);
	if (size2>0) 
	{
		*(File->Pos) += size2;
		WorkDone++;
		
		if (size2 == size) /* All of the page is sent */
		{
			page_cache_release(File->page);
			File->page = NULL;
		}
	}
	    	
	/* Step 7: Perform readahead if needed */
/*	printk("TryFile: Step 7 \n");*/
	
	/* Not implemented for now */
	
	
	return WorkDone;
	
	
HaveError:
	page_cache_release(File->page);
	
	return 0;
}

#endif