/*

kHTTPd -- the next generation

Send actual file-data to the connections

*/
/****************************************************************
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************/

/*

Purpose:

Userspace() hands all requests in the queue to the userspace-daemon, if
such beast exists.

Return value:
	The number of requests that changed status
*/
#include <linux/kernel.h>

#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/net.h>
#include <linux/sched.h>
#include <linux/skbuff.h>
#include <linux/smp_lock.h>
#include <linux/un.h>
#include <linux/unistd.h>

#include <net/ip.h>
#include <net/sock.h>
#include <net/tcp.h>

#include <asm/atomic.h>
#include <asm/semaphore.h>
#include <asm/processor.h>
#include <asm/uaccess.h>

#include <linux/file.h>


#include "structure.h"
#include "prototypes.h"
#include "sysctl.h"

int Userspace(int CPU)
{
	struct http_request *CurrentRequest,**Prev,*Next;
	
	CurrentRequest = UserspaceQueue[CPU];
	Prev = &(UserspaceQueue[CPU]);
	
	while (CurrentRequest!=NULL)
	{
#ifndef BENCHMARK	
		if (AddSocketToAcceptQue(CurrentRequest->sock,sysctl_khttpd_clientport)>=0)
		{
			(*Prev) = CurrentRequest->Next;
			Next = CurrentRequest->Next;
			
			if (CurrentRequest->sock!=NULL)
			{
				remove_wait_queue(CurrentRequest->sock->sk->sleep,&(CurrentRequest->sleep));
			}
			CurrentRequest->sock = NULL;	/* We no longer own it */
			
			CleanUpRequest(CurrentRequest); 
				
			CurrentRequest = Next;
			continue;
		
		}
		else
#endif		
		{
			(*Prev) = CurrentRequest->Next;
			Next = CurrentRequest->Next;
			
			
			Send403(CurrentRequest->sock);
			
			CleanUpRequest(CurrentRequest); 
				
			CurrentRequest = Next;
			continue;
		
		}
		
		

		Prev = &(CurrentRequest->Next);	
		CurrentRequest = CurrentRequest->Next;
	}
	return 0;
}

void StopUserspace(int CPU)
{
	struct http_request *CurrentRequest;
	
	CurrentRequest = UserspaceQueue[CPU];

	while (CurrentRequest!=NULL)
	{
		CleanUpRequest(CurrentRequest);
		CurrentRequest=CurrentRequest->Next;		
		schedule();
	}
}

extern struct sock *tcp_v4_lookup_listener(u32 daddr, unsigned short hnum, int dif);


struct sock *FindUserspace(unsigned short Port)
{
	return tcp_v4_lookup_listener(INADDR_ANY,Port,0);
}

void dummy_destructor(struct open_request *req)
{
}

struct or_calltable Dummy = 
{
 	NULL,
 	&dummy_destructor,
 	NULL
 };

int AddSocketToAcceptQue(struct socket *sock,int Port)
{
	struct open_request *req;
	struct sock *sk;
	struct tcp_opt *tp;
	
	sk = FindUserspace(Port);	
	
	if (sk==NULL) 
	{
		return -1;
	}
	req = tcp_openreq_alloc();
	if (req==NULL)
		return -1;	
	
	req->sk		= sock->sk;
	if (req->sk==NULL)
	 printk("WOOPS NULL \n");
	req->class	= &Dummy;

	lock_sock(sk);
	
	tp =&(sk->tp_pinfo.af_tcp);
	tcp_synq_queue(tp,req);	
	sk->ack_backlog++;
	tcp_inc_slow_timer(TCP_SLT_SYNACK);
	
	release_sock(sk);
	
	wake_up(sk->sleep);
	
	
  /* FIXME!: Now find a way to free the socket without freeing the sock */

	if (sock->fasync_list)
		printk(KERN_ERR "sock_release: fasync list not empty!\n");

/*	--sockets_in_use;*/	/*  NOT AVAILABLE TO MODULES!!!*/
	sock->file=NULL;
	iput(sock->inode);
	
	return 0;	
	
	
	
}

