/*

kHTTPd -- the next generation

Wait for headers on the accepted connections

*/
/****************************************************************
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************/

/*

Purpose:

WaitForHeaders polls all connections in "WaitForHeaderQueue" to see if
headers have arived. If so, the headers are decoded and the request is
moved to either the "SendingDataQueue" or the "UserspaceQueue".

Return value:
	The number of requests that changed status
*/

#include <linux/kernel.h>
#include <linux/skbuff.h>

#include <asm/uaccess.h>

#include "structure.h"
#include "prototypes.h"
#include "config.h"

static	char			*Buffer[KHTTPD_NUMCPU];


int WaitForHeaders(int CPU)
{
	struct http_request *CurrentRequest,**Prev;
	struct sock *sk;
	int count = 0;
	
	CurrentRequest = WaitForHeaderQueue[CPU];
	
	Prev = &(WaitForHeaderQueue[CPU]);
	
	while (CurrentRequest!=NULL)
	{
		
		/* If the connection is lost, remove from queue */
		
		if (CurrentRequest->sock->sk->state!=TCP_ESTABLISHED)
		{
			struct http_request *Next;
			
			Next = CurrentRequest->Next;
			
			*Prev = CurrentRequest->Next;
			CurrentRequest->Next = NULL;
		
			CleanUpRequest(CurrentRequest);
			CurrentRequest = Next;
			continue;
		}
		
		
		
		/* If data pending, take action */	
		
		sk = CurrentRequest->sock->sk;
		
		
		if (atomic_read(&(sk->rmem_alloc))>0) /* Do we have data ? */
		{
			struct http_request *Next;
			
			/* Decode header */
			
			if (DecodeHeader(CPU,CurrentRequest)<0)
			{
				CurrentRequest = CurrentRequest->Next;
				continue;
			} 
			
			
			/* Remove from WaitForHeaderQueue */		
			
			Next= CurrentRequest->Next;
		
			*Prev = Next;
			count++;
			
			/* Add to either the UserspaceQueue or the DataSendingQueue */
			
			if (CurrentRequest->IsForUserspace)
			{
				CurrentRequest->Next = UserspaceQueue[CPU];
				UserspaceQueue[CPU] = CurrentRequest;	
			} else
			{
				CurrentRequest->Next = DataSendingQueue[CPU];
				DataSendingQueue[CPU] = CurrentRequest;	
				ReadRest(CurrentRequest->sock);
#ifdef SENDFILE				
				/* Add to async queue */
				{
					struct async_sendfile *ASYNC;
					
					ASYNC = kmalloc(sizeof(struct async_sendfile),GFP_KERNEL);
					memset(ASYNC,0,sizeof(struct async_sendfile));
					ASYNC->filp = CurrentRequest->filp;
					ASYNC->sock = CurrentRequest->sock;
					ASYNC->Pos  = &(CurrentRequest->BytesSent);
					ASYNC->Length= &(CurrentRequest->FileLength);
					
					AddSendFile(ASYNC,&(CurrentRequest->sleep));
					
				}
#endif				
			}	
			
			CurrentRequest = Next;
			continue;
		
		}
		
		Prev = &(CurrentRequest->Next);
		CurrentRequest = CurrentRequest->Next;
	}
/*	if (count>0) printk("WaitHeaders : %i \n",count);*/
	return count;
}

void StopWaitingForHeaders(int CPU)
{
	struct http_request *CurrentRequest;
	
	CurrentRequest = WaitForHeaderQueue[CPU];

	while (CurrentRequest!=NULL)
	{
		CleanUpRequest(CurrentRequest);
		CurrentRequest->sock = NULL;
		CurrentRequest=CurrentRequest->Next;		
	}
	free_page((unsigned long)Buffer[CPU]);
}



int DecodeHeader(int CPU, struct http_request *Request)
{
	struct msghdr		msg;
	struct iovec		iov;
	int			len;

	mm_segment_t		oldfs;
	
	
	/* First, read the data */
	
	msg.msg_name     = 0;
	msg.msg_namelen  = 0;
	msg.msg_iov	 = &iov;
	msg.msg_iovlen   = 1;
	msg.msg_control  = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags    = 0;
	
	iov.iov_base = &Buffer[CPU][0];
	iov.iov_len  = (size_t)4096;
	
	len = 0;

	oldfs = get_fs(); set_fs(KERNEL_DS);
	len = sock_recvmsg(Request->sock,&msg,4096,MSG_PEEK);
	set_fs(oldfs);
	
	/* Then, decode the header */
	
	ParseHeader(Buffer[CPU],len,Request);
	
	Request->filp = OpenFileForSecurity(Request->FileName);
	
	if (Request->filp==NULL)
	{
		Request->IsForUserspace = 1;
		return 0;
	}
	else
	if ((Request->filp->f_dentry!=NULL)&&(Request->filp->f_dentry->d_inode!=NULL))
	{
		Request->FileLength = Request->filp->f_dentry->d_inode->i_size;
		Request->Time       = Request->filp->f_dentry->d_inode->i_mtime;
		Request->IMS_Time   = mimeTime_to_UnixTime(Request->IMS);
		

		Request->sock->sk->nonagle = 2;
		/*
		HTTPHeader(Request->FileName,Request->FileLength,Request->Time,Request->Header);
		SendBuffer(Request->sock,Request->Header,strlen(Request->Header));
		*/
		if (Request->IMS_Time>Request->Time)
		{	/* Not modified since last time */
			Send304(Request->sock);
			Request->FileLength=0;
		}
		else   /* Normal Case */
			SendHTTPHeader(Request);
		
	
	} else 
	{
		return -1;	
	};
	return 0;
}


int InitWaitHeaders()
{
	int I,I2;

	I=0;	
	while (I<KHTTPD_NUMCPU)
	{
		Buffer[I] = (char*)get_free_page(GFP_KERNEL);
		if (Buffer[I] == NULL) 
		{
			I2=0;
			while (I2<KHTTPD_NUMCPU-1)
			{
				free_page( (unsigned long)Buffer[I2++]);
			}
			return -1;
		};
		I++;
	}
		
	return 0;

}